/****************************************************************/
/**@file PCA7428.c:
 * Description: API layer for TEDIA PCA-7428 card.              *
 * Dependency: Windows 32, Windows 64 or Linux                  *
 *                Copyright 2014-2020 Jaroslav Fojtik           *
 ****************************************************************/

#if defined(_WIN32) || defined(_WIN64)
  #include <windows.h>
#else
  #include <sys/io.h>
#endif
#include <malloc.h>
#include <math.h>

#include "hudaqlib.h"
#include "hudaq_internal.h"


/** Ranges for both Analog Inputs and Analog Outputs. */
static const HudaqRange PCA7428RangeAIO[3] =
        {{0,+5},{-5,+5},{0,+10}};


#define DI_CHANNELS   2		///< Amount of Digital Outputs
#define DO_CHANNELS   1		///< Amount of Digital Inputs
#define DA_CHANNELS   2         ///< Amount of Analog Outputs
#define AD_CHANNELS   8         ///< Amount of Analog Inputs
#define ENC_CHANNELS  0         ///< Amount of Encoders
#define CTR_CHANNELS  1		///< Amount of counters
#define STEP_CHANNELS 0         ///< Amount of steppers


#define MASTERFREQUENCY 1000        ///< Internal oscillator frequency is 1kHz


/** Configuration structure for one analog output. */
typedef struct
{
  unsigned __int8 LoCache;
  unsigned __int8 HiCache;

  unsigned int Raw:1;           ///< 0-volts; 1-raw
} DASetup;



/** Cache of PCA-7428 state */
typedef struct
{
  UserDataHeader Hdr;                   ///< General device setup

  unsigned __int8 DOutReg;		///< Port configuration register
  //unsigned __int8 DIOCfgReg;
  unsigned __int8 IRQCfgReg;
  unsigned __int8 INTEnReg;  
  unsigned __int8 DACRangeReg;		///< Cached read only
  unsigned __int8 CWReg;		///< Commanding redister of scan circuits
  unsigned __int8 TimerReg;

  DASetup DAOptions[DA_CHANNELS];       ///< DA values to be cached
  unsigned __int8 ADRaw[AD_CHANNELS];
  DWORD ScanData[8];
  DWORD ScanData192;			///< Last scanning parameter.
} PCA_7428_Private;


typedef enum
{
  SCAN_OFF	= 0,
  SCAN_SW	= 1,
  SCAN_TIMER	= 2,
  SCAN_EXTERN	= 3,
  SCAN_CONTINUOUS= 5
} E_SCAN_MODE;


/** Symbolic aliases for available registers inside BADR4 read. */
typedef enum
{
  DInReg     = 0,
  DOutReg    = 1,	//4
  DInExtReg  = 2,	//8,

  Dac0RegLo	= 0x40/4,	// 40h, 44h
  Dac0RegHi	= 0x44/4,
  Dac1RegLo	= 0x48/4,	// 48h, 4Ch
  Dac1RegHi	= 0x4C/4,

  //DIOCfgReg  = 0x80/4,

  IRQCfgReg  =	0x180/4,	// WR only
  IRQStatusReg =0x180/4,	// RD only
  IRQClrReg =	0x184/4,	// WR only

  INTEnReg =	0x18C/4,	// RW

  CWReg =	0x1C0/4,	// WR only
  SWTrigReg =	0x1C4/4,	// WR only
  SWFIFODataReg=0x1C8/4,	// RO
  ScanAdrReg =	0x1E8/4,
  ScanDataReg0= 0x1F0/4,
  ScanDataReg1= 0x1F4/4,
  ScanDataReg2= 0x1F8/4,
  ScanDataReg3= 0x1FC/4,

  DACRangeReg = 0x3D0/4,	// RO
  TimerReg =	0x3F0/4,	// RW

} PCA7428_MEM1IO;

#define SWTrigStatusReg SWTrigReg


static __int8 Ports[DI_CHANNELS] = {DInReg, DInExtReg};


/** write to memory mapped device byte wise, bytes are stored on every 4th position. */
static __inline void StoreByte(size_t Ptr, int Offset, unsigned __int8 value)
{
  ((volatile unsigned __int8 *)(Ptr))[4*Offset] = value;
}

/** write to memory mapped device byte wise through cache. */
static __inline void StoreCachedByte(size_t Ptr, int Offset, unsigned __int8 value, unsigned __int8 *CachedValue)
{
  if (*CachedValue != value)
    {
    *CachedValue = value;
    StoreByte(Ptr,Offset,value);
    }
}


/** This inline is used for reading tightly packed OX9162 local configuration registers. */
static __inline unsigned __int8 GetBytePlain(size_t Ptr, int Offset)
{
  return ((volatile unsigned __int8 *)(Ptr))[4*Offset];
}


/** Convert voltage to a raw values for D/A convertor. */
static __inline __int16 PCA7428_VOLT2DA(double y, const HudaqRange *pRange)
{
  y = 0xFFFF*(y-pRange->Lo) / (pRange->Hi-pRange->Lo);
  if (y>=0xFFFF) return(0xFFFF);
  if (y<=0) return(0x0000);
  return((__int16)y);
}


static __inline double PCA7428_AI2VOLT(unsigned __int16 y)
{
  return (20)*(y/65535.0) - 10.0;
}



/****************************************************************
 *                                                              *
 *                 DIGITAL INPUTS & OUTPUTS                     *
 *                                                              *
 ****************************************************************/


/** Get data from digital input. */
static int PCA7428DIRead(const DeviceRecord *DevRecord, unsigned channel)
{
  if (channel>=DI_CHANNELS) return HUDAQBADARG;
  return GetBytePlain(DevRecord->DrvRes.Resources.MemResources[0].Base,Ports[channel]);
}


/** Write data to digital output. */
static HUDAQSTATUS PCA7428DOWrite(const DeviceRecord *DevRecord, unsigned channel, unsigned value)
{
  if (channel>=DO_CHANNELS) return HUDAQBADARG;
  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, DOutReg, value,
                  &((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->DOutReg);
  return HUDAQSUCCESS;
}


static HUDAQSTATUS PCA7428DIReadMultiple(const DeviceRecord *DevRecord, unsigned number, const unsigned *channels, unsigned *values)
{
unsigned FlagX = 0;
unsigned i;
size_t Base;
unsigned __int8 Cache[DI_CHANNELS+1];  //allocate more bytes for padding ind

  if (number==0 || channels==NULL || values==NULL) return HUDAQBADARG;
  if(DevRecord->pCT->HudaqDIRead==NULL) return HUDAQFAILURE;

  Base = DevRecord->DrvRes.Resources.MemResources[0].Base;

	/* Schedule what to read. */
  for(i=0; i<number; i++)
  {
    if (channels[i]<DI_CHANNELS)
      {
      FlagX |= 1<<(channels[i]);
      }
    else
      {
      values[i] = 0;
      FlagX |= 0x8000;
      }
    }
    
    /* Read isolated DIs */
  for(i=0; i<DI_CHANNELS; i++)
    {
    if(FlagX & (1<<i))
      {
      Cache[i] = GetBytePlain(Base,Ports[i]);
      }
    }  
    
	/* Final store of numbers read. */
  for(i=0; i<number; i++)
  {
    if(channels[i]<DI_CHANNELS)
      {
      values[i] = Cache[channels[i]];
      }
  }      
  
  if(FlagX == 0x8000) return HUDAQFAILURE;	//no valid value returned
  if(FlagX & 0x8000) return HUDAQPARTIAL;	//several values are valid and several are not
  return HUDAQSUCCESS;			//all values are valid (or no channel asked)
}


/** Write one bit to digital output. */
static void PCA7428DOWriteBit(const DeviceRecord *DevRecord, unsigned channel, unsigned bit, int value)
{
__int8 DoutByte;

  if(channel>=DO_CHANNELS) return;
  if(bit>=8) return;		/* There are only 8 bits available in PCA7428. */

  DoutByte = (((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->DOutReg);
  if(value) DoutByte |= (1<<bit);
       else DoutByte &= ~(1<<bit);

  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, DOutReg, DoutByte,
                  &((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->DOutReg);
}


static void PCA7428DOWriteMultipleBits(const DeviceRecord *DevRecord, unsigned channel, unsigned mask, unsigned value)
{
__int8 DoutByte;

  if(channel>=DO_CHANNELS) return;

  DoutByte = (((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->DOutReg);
  DoutByte = (DoutByte & ~mask) | (value & mask);

  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, DOutReg, DoutByte,
                  & ((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->DOutReg);
}


static double PCA7428DIOGetParameter(unsigned channel, HudaqParameter param)
{
  switch((int)param)
    {
    case HudaqDINUMBITS:
           return (channel<DI_CHANNELS) ? 8 : WRONG_VALUE;
    case HudaqDONUMBITS:
           return (channel<DO_CHANNELS) ? 8 : WRONG_VALUE;
    case HudaqDINUMCHANNELS:
	   return DI_CHANNELS;
    case HudaqDONUMCHANNELS:
	   return DO_CHANNELS;
    }

  return WRONG_VALUE;
}


/*
static HUDAQSTATUS PCA7428DIOSetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param, int value)
{
unsigned __int8 chval;

  switch((int)param)
    {
    case HudaqDINUMBITS:
           if(channel>=DI_CHANNELS) return WRONG_VALUE;
           return (value==8) ? 0 : WRONG_VALUE;
    case HudaqDONUMBITS:
           if(channel>=DI_CHANNELS) return WRONG_VALUE;
           return (value==8) ? 0 : WRONG_VALUE;
    case HudaqDINUMCHANNELS:
	   return DI_CHANNELS;
    case HudaqDONUMCHANNELS:
	   return DO_CHANNELS;
    case HudaqDOMODE:
          if(channel>=DO_CHANNELS) return WRONG_VALUE;
          //chval = ((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->DIOCfgReg;
	  //printf("chval =  %d\n",chval);          
          switch(channel)
          {
            case 0: if(value==1) chval &= ~0x1;
        	    if(value==0) chval |= 0x1;
        	    break;
            case 1: if(value==1) chval &= ~0x02;
        	    if(value==0) chval |= 0x02;
        	    break;
            case 2: if(value==1) chval &= ~0x04;
        	    if(value==0) chval |= 0x04;
		    break;
            case 3: if(value==1) chval &= ~0x08;
        	    if(value==0) chval |= 0x08;
        	    break;
          }
          StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, DIOCfgReg, chval,
                  & ((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->DIOCfgReg );          
          return HUDAQSUCCESS;
    }

  return WRONG_VALUE;
}
*/


/****************************************************************
 *                                                              *
 *                COUNTERS PWM + Count + Step                   *
 *                                                              *
 ****************************************************************/

static HUDAQSTATUS PCA7428PWMWrite(const DeviceRecord *DevRecord, unsigned channel, double frequency, double dutycycle)
{
double T;
HUDAQSTATUS ret = HUDAQSUCCESS;

  if (channel>=CTR_CHANNELS) return HUDAQBADARG;

  if (frequency<=0) return HUDAQBADARG; //Allow duty cycle 0 and 1 for f=0.

  T = MASTERFREQUENCY/frequency;
  if(T<1) 
  {
    StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 1,
                  & ((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);
    return HUDAQPARTIAL;
  }
  if(T>255)
  {
    StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 255,
                  & ((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);
    return HUDAQPARTIAL;
  }

  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 
		  (unsigned __int8)T,
                  &((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);

  return HUDAQSUCCESS;
}


static HUDAQSTATUS PCA7428CtrReset(const DeviceRecord *DevRecord, unsigned channel)
{
  if (channel>=CTR_CHANNELS) return HUDAQBADARG;
  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 
		  0, &((PCA_7428_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);
  return HUDAQSUCCESS;
}


static int PCA7428CtrRead(const DeviceRecord *DevRecord, unsigned channel)
{
  if (channel>=CTR_CHANNELS) return 0;
  return GetBytePlain(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg);
}


/****************************************************************
 *                                                              *
 *                  ANALOG INPUTS & OUTPUTS                     *
 *                                                              *
 ****************************************************************/


static double PCA7428AIGetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param)
{
PCA_7428_Private *Cache;

  if (channel>=AD_CHANNELS) return WRONG_VALUE;
  Cache = (PCA_7428_Private*)DevRecord->DrvRes.DriverData;

  switch(param)
    {
    case HudaqAIUNITS:
           return(Cache->ADRaw[channel]);

    case HudaqAIRange:
           return 3;	// (05V, +-5V, 010V) >>+-10V<<

    case HudaqAINUMCHANNELS:
	   return AD_CHANNELS;
    }

  return WRONG_VALUE;
}


/** Get data from analog input. */
static double PCA7428AIRead(const DeviceRecord *DevRecord, unsigned channel)
{
PCA_7428_Private *Cache;
size_t Ptr1;
unsigned i;

  if(channel>=AD_CHANNELS) return UNDEFINED_VALUE;

  Cache = (PCA_7428_Private *)DevRecord->DrvRes.DriverData;
  Ptr1 = DevRecord->DrvRes.Resources.MemResources[0].Base;

  if(Cache->ScanData[0] != (0x0A000000|channel))
  {
    StoreCachedByte(Ptr1, CWReg, 0, &Cache->CWReg);  // scanning must be turned off to access this
    StoreByte(Ptr1, ScanAdrReg, 0);	// Set address to 0

    StoreByte(Ptr1, ScanDataReg0, channel);
    StoreByte(Ptr1, ScanDataReg1, 0);	// AIN
    StoreByte(Ptr1, ScanDataReg2, 0);	// Gain 1x
    StoreByte(Ptr1, ScanDataReg3, 10);	// 10us basic (2us MUX not needed for 8 chans)
    Cache->ScanData[0] = 0x0A000000|channel;
  }

  if(Cache->ScanData192 != 0)		// Last scanning parameter.
  {
    StoreCachedByte(Ptr1, CWReg, 0, &Cache->CWReg);  // scanning must be turned off to access this
    StoreByte(Ptr1, ScanAdrReg, 192);	// Set address to 192
    StoreByte(Ptr1, ScanDataReg0, 0);
    StoreByte(Ptr1, ScanDataReg1, 0);
    StoreByte(Ptr1, ScanDataReg2, 0);
    StoreByte(Ptr1, ScanDataReg3, 0);
    Cache->ScanData192 = 0;
  }

  StoreCachedByte(Ptr1, CWReg, 1, &Cache->CWReg);	// Software trigger on.

  StoreByte(Ptr1, SWTrigReg, 1);	// Trig conversion.
  i = 0;
  while((GetBytePlain(Ptr1,SWTrigStatusReg) & 1) != 0)
  {    
    if(++i > 30) return UNDEFINED_VALUE;  // 9 loops are usually needed.
  }

  i = GetBytePlain(Ptr1,SWFIFODataReg);
  i |= 256*GetBytePlain(Ptr1,SWFIFODataReg);

  //printf("(%u:%u)", channel,i);

  if(Cache->ADRaw[channel])
    return i;
  else
    return PCA7428_AI2VOLT(i);
}


static HUDAQSTATUS PCA7428AIReadMultiple(const DeviceRecord *DevRecord, unsigned number, const unsigned *channels, double *values)
{
unsigned __int16 Mask;
unsigned __int16 address, realaddr;
int i;
size_t Ptr1;
HUDAQSTATUS ret = HUDAQSUCCESS;
PCA_7428_Private *Cache;

  if(number<=0 || channels==NULL || values==NULL) return HUDAQBADARG;

  Mask = 0;       // Prepare list of channels to convert.
  for(i=0; i<number; i++)
  {
    values[i] = UNDEFINED_VALUE;
    if(channels[i] < AD_CHANNELS)
      Mask |= 1<<channels[i];
    else
      ret = HUDAQPARTIAL;
  }

  Cache = (PCA_7428_Private *)DevRecord->DrvRes.DriverData;
  Ptr1 = DevRecord->DrvRes.Resources.MemResources[0].Base;

  realaddr = 0xFFFF;
  address = 0;
  for(i=0; i<AD_CHANNELS;i++)
  {
    if ( (Mask & (1<<i)) != 0)
    {
      if(Cache->ScanData[address] != (0x0A000000|i))
      {
        StoreCachedByte(Ptr1, CWReg, 0, &Cache->CWReg);  // scanning must be turned off to access this
        if(address != realaddr)
        {
          StoreByte(Ptr1, ScanAdrReg, address);	// Set address to 0
          realaddr = address;
        }

        StoreByte(Ptr1, ScanDataReg0, i);
        StoreByte(Ptr1, ScanDataReg1, 0);	// AIN
        StoreByte(Ptr1, ScanDataReg2, 0);	// Gain 1x
        StoreByte(Ptr1, ScanDataReg3, 10);	// 10us basic (2us MUX not needed for 8 chans)
        Cache->ScanData[address] = 0x0A000000|i;
        realaddr++;
      }
      address++;
    }
  }

  address--;
  if(Cache->ScanData192 != address)	// Last scanning parameter.
  {
    StoreCachedByte(Ptr1, CWReg, 0, &Cache->CWReg);  // scanning must be turned off to access this
    StoreByte(Ptr1, ScanAdrReg, 192);	// Set address to 192
    StoreByte(Ptr1, ScanDataReg0, address);
    StoreByte(Ptr1, ScanDataReg1, 0);
    StoreByte(Ptr1, ScanDataReg2, 0);
    StoreByte(Ptr1, ScanDataReg3, 0);
    Cache->ScanData192 = address;
  }
  address++;

  StoreCachedByte(Ptr1, CWReg, 1, &Cache->CWReg);	// Software trigger on.

  StoreByte(Ptr1, SWTrigReg, 1);	// Trig conversion.
  i = 0;
  while((GetBytePlain(Ptr1,SWTrigStatusReg) & 1) != 0)
  {    
    if(++i > address*12+10)
    {    
      return HUDAQFAILURE;  // 9 loops are usually needed per one channel.
    }
  }

  for(address=0; address<=Cache->ScanData192; address++)
  {
    Mask = GetBytePlain(Ptr1,SWFIFODataReg);
    Mask |= 256*GetBytePlain(Ptr1,SWFIFODataReg);

    for(i=0; i<number; i++)
    {
      if(channels[i] == (Cache->ScanData[address] & 0xFF))
      {
        if(Cache->ADRaw[channels[i]])
          values[i] = Mask;
        else
          values[i] = PCA7428_AI2VOLT(Mask);
      }
    }
  }

  return ret;
}


static double PCA7428AOGetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param)
{
PCA_7428_Private *Cache;

  Cache = (PCA_7428_Private*)DevRecord->DrvRes.DriverData;
  if(channel>=DA_CHANNELS) return WRONG_VALUE;

  switch(param)
    {
    case HudaqAOUNITS:
           return(Cache->DAOptions[channel].Raw);
    case HudaqAORange:           
           return (Cache->DACRangeReg>>2*channel) & 3;		// (05V, +-5V, 010V)
    case HudaqAONUMCHANNELS:
	   return DA_CHANNELS;
    }

  return WRONG_VALUE;
}


/** Write data to analog output. */
static void PCA7428AOWrite(const DeviceRecord *DevRecord, unsigned channel, double value)
{
PCA_7428_Private *Cache;
WORD OutputRaw;
size_t Ptr0;

  if (channel>=DA_CHANNELS) return;
  Ptr0 = DevRecord->DrvRes.Resources.MemResources[0].Base;
  if(Ptr0==0) return;

  Cache = (PCA_7428_Private*)DevRecord->DrvRes.DriverData;
  if(Cache->DAOptions[channel].Raw==1 || (Cache->DACRangeReg&3)==3)
      OutputRaw = (WORD)value;
  else
      OutputRaw = PCA7428_VOLT2DA(value, &PCA7428RangeAIO[(Cache->DACRangeReg>>2*channel)&3]);
   
  if((OutputRaw&0xFF)==Cache->DAOptions[channel].LoCache && (OutputRaw>>8)==Cache->DAOptions[channel].HiCache)
      return;

  if(channel==0)
  {
    StoreCachedByte(Ptr0, Dac0RegLo, OutputRaw&0xFF, &Cache->DAOptions[0].LoCache);
    StoreByte(Ptr0, Dac0RegHi, OutputRaw>>8);
    Cache->DAOptions[0].HiCache = OutputRaw>>8;
  }
  else
  {
    StoreCachedByte(Ptr0, Dac1RegLo, OutputRaw&0xFF, &Cache->DAOptions[1].LoCache);
    StoreByte(Ptr0, Dac1RegHi, OutputRaw>>8);
    Cache->DAOptions[1].HiCache = OutputRaw>>8;
  }
}


/** Write multiple data to analog output synchronized. */
static HUDAQSTATUS PCA7428AOWriteMultiple(const DeviceRecord *DevRecord, unsigned number, const unsigned* channels, const double* values)
{
char Flag = 0;

  if (channels==NULL || values==NULL || number<=0) return HUDAQBADARG;

  while(number-->0)
  {
    if(*channels>=DA_CHANNELS)
    {
      Flag |= 2;
    }  
    else
    {
      PCA7428AOWrite(DevRecord, *channels, *values);
      Flag |= 1;      
    }
    values++;
    channels++;
  }

  if(Flag==3) return HUDAQPARTIAL;	//several values are valid and several are not
  if(Flag==2) return HUDAQFAILURE;	//no valid value returned
  return HUDAQSUCCESS;			//all values are valid (or no channel asked)
}




/********************************************************************/

/** Initialization procedure for PCA-7428x cards. */
static int InitPCA7428(DeviceRecord *DevRecord, int IniOptions)
{
PCA_7428_Private *Cache;
size_t Base;

  if (DevRecord==NULL) return -1;

  if (DevRecord->DrvRes.Resources.NumMemResources<2) return -2;  //insufficient amount of resources
  Cache = DevRecord->DrvRes.DriverData;
  if (Cache==NULL) return -3;
  if (DevRecord->DrvRes.DriverDataSize<sizeof(PCA_7428_Private)) return -4;

  DevRecord->pCT = &CTPCA7428CL;

  if ((IniOptions & HudaqOpenNOINIT)==0)
  {                           //initialize hardware only if no other application is running
    Base = DevRecord->DrvRes.Resources.MemResources[0].Base;

	/* disable interrupts */
    Cache->INTEnReg = 0;
    StoreByte(Base, INTEnReg, 0);

      /* Initialization of DI and DO. */
    Cache->DOutReg = 0;
    StoreByte(Base, DOutReg, 0);

//    Cache->DIOCfgReg = 0;
//   StoreByte(Base, DIOCfgReg, 0);

    Cache->IRQCfgReg = 0;
    StoreByte(Base, IRQCfgReg, 0);

    Cache->TimerReg = 0;
    StoreByte(Base, TimerReg, 0);	/* Stop timer. */

    Cache->DACRangeReg = GetBytePlain(DevRecord->DrvRes.Resources.MemResources[0].Base, DACRangeReg);

      /* Initialization of AO. */
    memset(Cache->DAOptions, 0, sizeof(Cache->DAOptions));
    if((Cache->DACRangeReg&3) == 1)
        Cache->DAOptions[0].HiCache = 0x80;  // Symetrical AO range; 0V is in the middle.

    if(((Cache->DACRangeReg>>2)&3) == 1)
        Cache->DAOptions[1].HiCache = 0x80;  // Symetrical AO range; 0V is in the middle.
    
    StoreByte(Base, Dac0RegLo, Cache->DAOptions[0].LoCache);
    StoreByte(Base, Dac0RegHi, Cache->DAOptions[0].HiCache);
    StoreByte(Base, Dac1RegLo, Cache->DAOptions[1].LoCache);
    StoreByte(Base, Dac1RegHi, Cache->DAOptions[1].HiCache);

      /* Initialization of AI. */
    memset(Cache->ADRaw, 0, sizeof(Cache->ADRaw));
    Cache->CWReg = 0;
    StoreByte(Base, CWReg, SCAN_OFF);		// Turn off ADC conversion.

    memset(Cache->ScanData, 0xFF, sizeof(Cache->ScanData));
    Cache->ScanData192 = 0xFFFFFFFF;
  }  

return 1; //success
}


/** Internal cleanup procedure for PCA7428 */
static void DonePCA7428(DeviceRecord *DevRecord)
{
  if (DevRecord==NULL) return;

  DevRecord->pCT = &CtDummy;
}


static const HudaqRange *PCA7428QueryRange(const DeviceRecord *DevRecord, HudaqSubsystem S, unsigned item)
{
  switch(S & HudaqSubsystemMASK)
    {
    case HudaqAI:  if(item>=3) return NULL;
                   return &PCA7428RangeAIO[item];
    case HudaqAO:  if(item>=3) return NULL;
                   return &PCA7428RangeAIO[item];
    }
  return NULL;
}



/****************************************************************
 *                                                              *
 *                    GET/SET PARAMETERS                        *
 *                                                              *
 ****************************************************************/


static HUDAQSTATUS PCA7428IRQSetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param, int value)
{
size_t Base;

  Base = DevRecord->DrvRes.Resources.MemResources[0].Base;

  switch((int)param)
    {
    case HudaqIRQ+0:
	{
	 if(value)
	 {
//	   StoreCachedByte(Base, IRQCfgReg, 1, &((PCA_7428_Private *)DevRecord->DrvRes.DriverData)->IRQCfgReg);
//	   StoreByte(Base, IRQClrReg, 0xFF);
//	   StoreCachedByte(Base, INTEnReg, 0x80, &((PCA_7428_Private *)DevRecord->DrvRes.DriverData)->INTEnReg);
         }
         else
	 { 
//	   StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, INTEnReg, 0,
//			   &((PCA_7428_Private *)DevRecord->DrvRes.DriverData)->INTEnReg);
//	   StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, IRQCfgReg, 0,
//			   &((PCA_7428_Private *)DevRecord->DrvRes.DriverData)->IRQCfgReg);
         }
        return HUDAQSUCCESS;
        }
    case HudaqIRQ+1:
           return HUDAQSUCCESS;
    }

  return WRONG_VALUE;
}


static double PCA7428IRQGetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param)
{
  switch((int)param)
    {
    case HudaqIrqNUMCOUNTER: 
	   return ((PCA_7428_Private *)DevRecord->DrvRes.DriverData)->Hdr.IRQcounter;
    //case HudaqIRQ+1:
    //       return GetBytePlain(DevRecord->DrvRes.Resources.MemResources[0].Base,IRQStatusReg);
    }
  return WRONG_VALUE;
}


static HUDAQSTATUS PCA7428SetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param, double value)
{
switch(param & HudaqSubsystemMASK)
  {
  case HudaqDI:
//  case HudaqDO:  return PCA7428DIOSetParameter(DevRecord,channel,param,(int)value);
//  case HudaqAI:  
//  case HudaqAO:  
//  case HudaqEnc: 
//  case HudaqPWM:
//  case HudaqCtr: 
//  case HudaqStep:
    case HudaqIRQ: return PCA7428IRQSetParameter(DevRecord,channel,param,(int)value);
  }
return HUDAQNOTSUPPORTED;
}

static double PCA7428GetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param)
{
switch(param & HudaqSubsystemMASK)
  {
  case HudaqDI:
  case HudaqDO:  return PCA7428DIOGetParameter(channel,param);
  case HudaqAI:  return PCA7428AIGetParameter(DevRecord,channel,param);
  case HudaqAO:  return PCA7428AOGetParameter(DevRecord,channel,param);
//case HudaqEnc:
//case HudaqPWM:
//case HudaqCtr:
//case HudaqStep:
  case HudaqIRQ: return PCA7428IRQGetParameter(DevRecord,channel,param);
  }
return WRONG_VALUE;
}


const CallTable CTPCA7428CL =
    {
    "PCA-7428CL", 0x1760, 0x0241,
    InitPCA7428,
    DonePCA7428,

    PCA7428SetParameter,
    PCA7428GetParameter,
    PCA7428QueryRange,

        // INITIALIZE DI callers
    PCA7428DIRead,
    GenericDIReadBit,           //Generic implementation
    PCA7428DIReadMultiple,
        // INITIALIZE DO callers
    PCA7428DOWrite,
    PCA7428DOWriteBit,
    PCA7428DOWriteMultipleBits,
    GenericDOWriteMultiple,
        // INITIALIZE AI callers
    PCA7428AIRead,
    PCA7428AIReadMultiple,
        // INITIALIZE AO callers
    PCA7428AOWrite,
    PCA7428AOWriteMultiple,
        // INITIALIZE Enc callers
    NULL,                       // Not available
    NULL,
        // INITIALIZE Ctr callers
    PCA7428CtrRead,
    PCA7428CtrReset,
        // INITIALIZE PWM callers
    PCA7428PWMWrite,            
    NULL,			// Not available
    NULL,
        // INITIALIZE Step callers
    NULL,                       // Not available
    };


static int InitPCA7428CS(DeviceRecord *DevRecord, int IniOptions)
{
  int ret = InitPCA7428(DevRecord, IniOptions);
  if(ret>=0) DevRecord->pCT = &CTPCA7428CS;
  return ret;
}

const CallTable CTPCA7428CS =
    {
    "PCA-7428CS", 0x1760, 0x0243,
    InitPCA7428CS,
    DonePCA7428,

    PCA7428SetParameter,
    PCA7428GetParameter,
    PCA7428QueryRange,

        // INITIALIZE DI callers
    PCA7428DIRead,
    GenericDIReadBit,           //Generic implementation
    PCA7428DIReadMultiple,
        // INITIALIZE DO callers
    PCA7428DOWrite,
    PCA7428DOWriteBit,
    PCA7428DOWriteMultipleBits,
    GenericDOWriteMultiple,
        // INITIALIZE AI callers
    PCA7428AIRead,
    PCA7428AIReadMultiple,	//GenericAIReadMultiple
        // INITIALIZE AO callers
    PCA7428AOWrite,
    PCA7428AOWriteMultiple,
        // INITIALIZE Enc callers
    NULL,                       // Not available
    NULL,
        // INITIALIZE Ctr callers
    PCA7428CtrRead,
    PCA7428CtrReset,
        // INITIALIZE PWM callers
    PCA7428PWMWrite,            
    NULL,			// Not available
    NULL,
        // INITIALIZE Step callers
    NULL,                       // Not available
    };


static int InitPCA7428CE(DeviceRecord *DevRecord, int IniOptions)
{
  int ret = InitPCA7428(DevRecord, IniOptions);
  if(ret>=0) DevRecord->pCT = &CTPCA7428CE;
  return ret;
}

const CallTable CTPCA7428CE =
    {
    "PCA-7428CE", 0x1760, 0x0245,
    InitPCA7428CS,
    DonePCA7428,

    PCA7428SetParameter,
    PCA7428GetParameter,
    PCA7428QueryRange,

        // INITIALIZE DI callers
    PCA7428DIRead,
    GenericDIReadBit,           //Generic implementation
    PCA7428DIReadMultiple,
        // INITIALIZE DO callers
    PCA7428DOWrite,
    PCA7428DOWriteBit,
    PCA7428DOWriteMultipleBits,
    GenericDOWriteMultiple,
        // INITIALIZE AI callers
    PCA7428AIRead,
    PCA7428AIReadMultiple,	//GenericAIReadMultiple
        // INITIALIZE AO callers
    PCA7428AOWrite,
    PCA7428AOWriteMultiple,
        // INITIALIZE Enc callers
    NULL,                       // Not available
    NULL,
        // INITIALIZE Ctr callers
    NULL, //PCA7428CtrRead,
    NULL, //PCA7428CtrReset,
        // INITIALIZE PWM callers
    NULL, //PCA7428PWMWrite,            
    NULL,			// Not available
    NULL,
        // INITIALIZE Step callers
    NULL,                       // Not available
    };

